/***************************************************************************
 *                                                                         *
 *    LIBDSK: General floppy and diskimage access library                  *
 *    Copyright (C) 2005  John Elliott <jce@seasip.demon.co.uk>            *
 *                                                                         *
 *    This library is free software; you can redistribute it and/or        *
 *    modify it under the terms of the GNU Library General Public          *
 *    License as published by the Free Software Foundation; either         *
 *    version 2 of the License, or (at your option) any later version.     *
 *                                                                         *
 *    This library is distributed in the hope that it will be useful,      *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    *
 *    Library General Public License for more details.                     *
 *                                                                         *
 *    You should have received a copy of the GNU Library General Public    *
 *    License along with this library; if not, write to the Free           *
 *    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,      *
 *    MA 02111-1307, USA                                                   *
 *                                                                         *
 ***************************************************************************/

/* DSKUTIL is the start of a console-mode disc editor, modelled after the 
 * CP/M utility DU90. Eventually it should allow interactive access to 
 * all of the features of LibDsk. At the moment all it lets you do is read 
 * arbitrary sectors. */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "libdsk.h"
#include "utilopts.h"
#include "formname.h"
#include <errno.h>

#ifdef CPM
#define AV0 "DSKUTIL"
#else
#define AV0 argv[0]
#endif

static dsk_format_t format = -1;
static	char *intyp = NULL;
static	char *incomp = NULL;
static	int inside = -1;
static  int idstep =  0;
static  int retries = 1;
static  dsk_ltrack_t track = 0;
static  dsk_psect_t sector = 0;
static int first = -1, last = -1;
static DSK_PDRIVER indr = NULL;
static DSK_GEOMETRY dg;
static unsigned char *secbuf;

int do_util(char *infile);
int obey(char *command);

int check_numeric(char *arg, int *argc, char **argv)
{
	int n = find_arg(arg, *argc, argv);
	unsigned nr;

	if (n < 0) return -1;	
	excise_arg(n, argc, argv);
	if (n >= *argc || atoi(argv[n]) == 0)
	{
		fprintf(stderr, "Syntax error: use '%s nnn' where nnn is nonzero\n", arg);
		exit(1);
	}
	nr = atoi(argv[n]);
	excise_arg(n, argc, argv);

	return nr;
}

static void report(const char *s)
{
        fprintf(stderr, "%s\r", s);
        fflush(stderr);
}

static void report_end(void)
{
        fprintf(stderr, "\r%-79.79s\r", "");
        fflush(stderr);
}


int help(int argc, char **argv)
{
	fprintf(stderr, "Syntax: \n"
                       "      %s {options} in-image { out-image }\n",
			AV0);
	fprintf(stderr,"\nOptions are:\n"
		       "-type <type>   type of input disc image\n"
                       "-side <side>   Force side 0 or side 1 of input\n"
		       "-retry <count> Set number of retries on error\n"
		       "-first <cyl>   Start at specified cylinder\n"
		       "-last <cyl>    Scan up to specified cylinder\n"
		       "-dstep         Double-step\n"
		       "-format        Force a specified format name\n");
	fprintf(stderr,"\nDefault type is autodetect.\n\n");
		
	fprintf(stderr, "eg: %s /dev/fd0\n"
                        "    %s -format pcw720 /dev/fd1\n",
			AV0, AV0);
	valid_formats();
	return 1;
}


int main(int argc, char **argv)
{

	if (find_arg("--version", argc, argv) > 0) return version(); 
	if (argc < 2) return help(argc, argv);
	if (find_arg("--help",    argc, argv) > 0) return help(argc, argv);

	ignore_arg("-itype", 2, &argc, argv);
	ignore_arg("-iside", 2, &argc, argv);
	ignore_arg("-idstep", 2, &argc, argv);
	ignore_arg("-icomp", 2, &argc, argv);
	ignore_arg("-otype", 2, &argc, argv);
	ignore_arg("-oside", 2, &argc, argv);
	ignore_arg("-oostep", 2, &argc, argv);
	ignore_arg("-ocomp", 2, &argc, argv);

	intyp     = check_type("-type", &argc, argv); 
	incomp    = check_type("-comp", &argc, argv); 
        inside    = check_forcehead("-side", &argc, argv);
	retries   = check_retry("-retry", &argc, argv);
	if (present_arg("-dstep", &argc, argv)) idstep = 1;
        format    = check_format("-format", &argc, argv);
	first     = check_numeric("-first", &argc, argv);
	last      = check_numeric("-last", &argc, argv);
 	args_complete(&argc, argv);
	return do_util(argv[1]);
}



static char *op = "Opening";


int do_util(char *infile)
{
	dsk_err_t e;
	int done = 0;
	char buffer[200], *c;

        dsk_reportfunc_set(report, report_end);
	e = dsk_open (&indr,  infile,  intyp, incomp);
	if (!e) e = dsk_set_retry(indr, retries);
	if (!e && inside >= 0) e = dsk_set_option(indr, "HEAD", inside);
	if (!e && idstep) e = dsk_set_option(indr, "DOUBLESTEP", 1);

/* Try to guess geometry. If this fails the user has to specify an approximate
 * geometry manually. We need to do this to get a feel for the number of 
 * cylinders to scan; dsk_psecid() will fail if (eg) it tries to select a
 * 1.4Mb geometry on a 360k drive */
        if (format == -1)
        {
                op = "Identifying disc";
                if (!e) e = dsk_getgeom(indr, &dg);
        }
        else if (!e) e = dg_stdformat(&dg, format, NULL, NULL);
	if (!e) 
	{
		secbuf = malloc(dg.dg_secsize);
		if (!secbuf) e = DSK_ERR_NOMEM;
	}
	if (e)
	{
		if (indr)  dsk_close(&indr);
		fprintf(stderr, "\n%s: %s\n", op, dsk_strerror(e));
		return 1;
	}
	do
	{
		printf(":");
		if (!fgets(buffer, sizeof(buffer), stdin))
		{
			strcpy(buffer, "x\n");
		}
		for (c = strtok(buffer, ";\n"); 
			c != NULL; 
			c = strtok(NULL, "\n"))
		{
			int e = DSK_ERR_OK;

			while (c[0] == ' ' || c[0] == '\t') ++c;
			if (c[0] == 'x' || c[0] == 'X') done = 1;
			else e = obey(c);
			if (e)
			{
				printf("Error in '%s': %s\n",
						c, dsk_strerror(e));
				break;
			}
		}
	} while (!done);
	if (indr)  dsk_close(&indr);
	if (secbuf) free(secbuf);
	if (e)
	{
		fprintf(stderr, "\n%s: %s\n", op, dsk_strerror(e));
		return 1;
	}
	return 0;
}

static void showts(void)
{
	printf("Track=%u Sector=%u\n", track, sector);
}

static void dump(void)
{
	int n, m;
	for (n = 0; n < dg.dg_secsize; n += 16)
	{
		printf("%04x: ", n);
		for (m = 0; m < 16; m++)
		{
			printf("%02x", secbuf[n+m]);
			if ((m & 3) == 3) putchar(' ');
		}
		putchar(' ');
		for (m = 0; m < 16; m++)
		{
			if (isprint(secbuf[n+m])) putchar(secbuf[n+m]);
			else			  putchar('.');
		}
		putchar('\n');
	}
}



int obey(char *cmd)
{
	dsk_pcyl_t cyl;
	dsk_phead_t head;

	switch(toupper(cmd[0]))
	{
		case 'D':
			dump();
			return DSK_ERR_OK;
		case 'T':
			if (isdigit(cmd[1])) track = atoi(cmd+1);
			showts();
			return DSK_ERR_OK;
		case 'S':
			if (isdigit(cmd[1])) sector = atoi(cmd+1);
			showts();
			return DSK_ERR_OK;
		case '-':
			--sector;
			if (sector < dg.dg_secbase || sector >= (dg.dg_sectors + dg.dg_secbase))
			{
				sector = dg.dg_secbase + dg.dg_sectors - 1;
				if (track > 0) --track;
			}
			showts();
			goto doread;
		case '+':
			++sector;
			if (sector >= (dg.dg_sectors + dg.dg_secbase))
			{
				sector = dg.dg_secbase;
				++track;
				dg_lt2pt(&dg, track, &cyl, &head);
				if (cyl >= dg.dg_cylinders || head >= dg.dg_heads) --track;
			}
			showts();
			/* FALL THROUGH */
		case 'R':
		doread:
			if (sector < dg.dg_secbase || 
			    sector >= (dg.dg_sectors + dg.dg_secbase))
			{
				printf("Cannot read: Sector not set or out of range\n");
				return DSK_ERR_BADVAL;
			}
			dg_lt2pt(&dg, track, &cyl, &head);
			return dsk_pread(indr, &dg, secbuf, cyl, head, sector);
		case 'J':
			if (sector < dg.dg_secbase || 
			    sector >= (dg.dg_sectors + dg.dg_secbase))
			{
				printf("Cannot read: Sector not set or out of range\n");
				return DSK_ERR_BADVAL;
			}
			dg_lt2pt(&dg, track, &cyl, &head);
			return dsk_pcheck(indr, &dg, secbuf, cyl, head, sector);
			
	}
	printf("Unknown command: '%s'\n", cmd);
	return DSK_ERR_OK;
}
